#!/usr/bin/perl -w

=head1 NAME

B<POWER::MetaSearch> - Make fast metasearch using many search servers.

=head1 SYNOPSIS

 use POWER::MetaSearch;

 @engines = EnginesList();
 %engines = %{ MetaSearchRaw("Programming Perl", 60 };
 %engines = %{ MetaSearchRaw("Programming Perl", 180, {
          'MSN'		=> undef,
          'AOL'		=> 50,
        } ) };

=head1 DESCRIPTION

Fast metasearch.
Procedures imported by default: MetaSearchRaw(), EnginesList().

=head1 KNOWN SEARCH SERVERS

Here the full list of search servers which can be used in metasearch:
'DirectHit', 'MSN', 'AOL', 'AllTheWeb', 'LycosPro', 'Excite',
'NorthernLight', 'Google', 'Yahoo', 'Netscape'.

=head1 FUNCTIONS

=over 4

=cut

package POWER::MetaSearch;
use POWER::lib;
use POWER::Multi::GET;
use POWER::HTML::TAG qw(strip);
use strict;
use vars qw($VERSION);
$VERSION = "1.01" ;

sub import {
    no strict;
    *{"${\scalar caller}::$_"} = \&{$_} for qw(MetaSearchRaw EnginesList);
}

# Define search servers
my %ENGINES = (
	"Google"	=> {
		url	=> "http://www.google.com/search?q=^~query~^&btnG=Google+Search",
		Rec	=> qr!<p><A HREF=([^>]+)>(.*?)</A>\s*(<font.*?)<font!si,
		Next	=> qr!<A HREF=(/search\?[^>]+)><IMG [^>]+><br><span [^>]+><b>Next!si,
	    },
	"AllTheWeb"	=> {
		url	=> "http://www.alltheweb.com/search?cat=web&lang=any&query=^~query~^",
		Rec	=> qr!<dt>.*?href=\"([^\"]*)\">(.*?)</dt>(.*?)<br>!si,
		Next	=> qr!<a class=\"link\" href=\"(search\?[^\"]+)\">\&nbsp;\&gt;\&gt;!si,
	    },
	"Excite"	=> {
		url	=> "http://search.excite.com/search.gw?c=web&search=^~query~^&onload=",
		Rec	=> qr!<li>\s+<a\s+href=/r/[^\s]*;([^\s]+).*?>(.*?)<br>(.*?)<br>!si,
		Next	=> qr!<a href=(http://search\.excite\.com/search\.gw\?[^>]+)><b>next!si,
	    },
	"AOL"		=> {
		url	=> "http://search.aol.com/dirsearch.adp?start=&from=topsearchbox.%2Findex.adp&query=^~query~^",
		Rec	=> qr!<li><b>.*?onmouseover=\"self\.status=\'([^\']+)\'[^>]*>(.*?)</A>(.*?)<br>!si,
		Next	=> qr!<A HREF=\"(dirsearch.adp\?[^\"]+)\"><b>more!si,
	    },
	"DirectHit"	=> {
		url	=> "http://www.directhit.com/search.asp?qry=^~query~^&alias=websrch&submit.x=0&submit.y=0",
		Rec	=> qr!\d\. <a href.*?url=(.*?)&.*?<b>(.*?)</a>(.*?)<span!si,
		Next	=> qr!<a href=\"(/umbrella.asp\?[^\"]+)\">Next &gt;&gt;!si,
	    },
	"LycosPro"	=> {
		url	=> "http://lycospro.lycos.com/srchpro/?aloc=sb_init_content&first=1&lpv=1&query=^~query~^&t=all&type=advwebsites",
		Rec	=> qr!<LI><FONT.*?&target=(.*?)&.*?>(.*?)</a>(.*?)</FONT>!si,
		Next	=> qr!<A HREF=(/srchpro/\?[^>]*)><B>next</B></A>!si,
	    },
	"MSN"		=> {
		url	=> "http://search.msn.com/results.asp?FORM=SMCRT&cfg=SMCINITIAL&q=^~query~^&un=doc&v=1",
		Rec	=> qr!<LI /><A CLASS=\"clsResult.*?href=\"(.*?)\">(.*?)</A>(.*?)<br /><!si,
		Next	=> qr!<a CLASS=\"clsNextPrev\"[^>]+href=\"(results.asp\?[^\"]*)\">NEXT!si,
	    },
	"Netscape"	=> {
		url	=> "http://search.netscape.com/search.psp?cp=nsikwphopNetscape&search=^~query~^", 
		Rec	=> qr!<LI><a href=\"http:[^\"]*?/nrpusrws/([^\"]*?)\">(.*?)</a>.*?<br>(.*?)<br>!si,
		Next	=> qr!<a href=\"(/search.psp\?[^\"]*)\">next!si,
	    },
	"NorthernLight"	=> {
		url	=> "http://www.northernlight.com/nlquery.fcg?cb=0&qr=^~query~^&search.x=0&search.y=0",
		Rec	=> qr!NLResultStart.*?<a href=\"([^\"]*)\">(.*?)</a>.*?</b>(.*?)<br>!si,
		Next	=> qr!<a href=\"(/nlquery.fcg\?[^\"]*)\"><img[^>]*Next Page!si,
	    },
	"Yahoo"		=> {
		url	=> "http://google.yahoo.com/bin/query?p=^~query~^&hc=6&hs=166",
		Rec	=> qr!<li>\s*<a href=\"http:[^\"]*?/\*([^\"]*)\">(.*?)</a>(.*?)<font!si,
		Next	=> qr!<a href=\"(/bin/query\?[^\"]*)\">Next!si,
	    },
);
$ENGINES{$_}{name} = $_ for keys %ENGINES;

=item B<EnginesList>()

Return sorted list of supported engines.

=cut

sub EnginesList { sort keys %ENGINES }

# Normalize data cutted from html page
sub norm { local $_ = UnEnc(strip($_[0])); s/\n+/ /g; s/^\s+//; s/\s+$//; $_ }
# Standart parsing: 1) list of urls with title and description, 2) link "Next"
sub ParseEngine {
    my ($html, %Engine, @results, $next_url) = ($_[0], %{$_[1]});
    push @results, {url => UnEncUri($1), title => norm($2), descr => norm($3)}
	while $html=~m!\G.*?$Engine{Rec}!gcsi;
    $next_url=abs_url($1, $Engine{url}) if $html=~m!\G.*?$Engine{Next}!gcsi;
    return (\@results, $next_url);
}

=item B<MetaSearchRaw>( $query, $timeout [,\%engines] )

Make metasearch and return all results without eliminate duplications and
without sorting.

$query is user query. No special query language used at this time: query
will be sent to search servers as is.

$timeout is timeout in seconds. metasearch finished after downloading asked
number of results for all asked servers, or after timeout.

%engines is a way to define which search servers must be used in metasearch,
and how much results must be downloaded. If %engines is absent or it has no 
keys ALL known search servers will be used with NO limit for maximum results.
%engines format is: ("server_name" => $max_results) [SEE SYNOPSIS]. 
If $max_results undefined no results limit will be used.

Return hash reference contains downloaded results for all asked servers.

 Format of returned data : 
 1) MetaSearchRaw() is returned hash reference. 
 2) Keys of this hash are search engines names, 
    values are reference to hash contains results from those engines.
 3) Those hashes has fields 'max_results' and 'results'. 
    'max_results' is set to $max_result or undef if $max_results was 
    not set for this engine.
    'results' is an array reference to array contains results returned 
    by engine.
 4) Each result is a hash reference and has 'url', 'title' and 'descr' 
    fields.

Example of result hash:

 {	"AOL"	=> {
		max_results	=> 50,
		results		=> [
			{
			    url	  => "http://url/first",
			    title => "The first title",
			    descr => "The first description ...",
			},
			{
			    url	  => "http://url/second",
			    title => "The second title",
			    descr => "The second description ...",
			},
		    ],
	    },
	"MSN"	=> {
		max_results	=> undef,
		results		=> [
			{
			    url	  => "http://url/AAA",
			    title => "The AAA title",
			    descr => "The AAA description ...",
			},
		    ],
	    },
 };

=cut

sub MetaSearchRaw {
    # %engines  : (_ => _, ...)
    my ($query, $timeout, %engines) = ($_[0], $_[1], %{$_[2]});
    #  %engines:   , 
    # _   {max_results}
    %engines = map {$_, {max_results=>$engines{$_}}} grep {! keys %engines ||
	exists $engines{$_}} keys %ENGINES;
    #       
    my %URL = map {
	(my $url=$ENGINES{$_}{url})=~s/\^~query~\^/EncUri($query)/eg;
	($ENGINES{$_}{name}." - 1", {url=>$url,engine=>$ENGINES{$_},from=>1})
	} keys %engines;
    $timeout = time()+$timeout;
    POWER::Multi::GET::Download(%URL, sub {
	    #       
	    for (grep {$_->{_ok}} values %URL) {
		#    (   timeout)
		delete $_->{_state}, next if $_->{_ok} eq "ERR";
		# DEBUG
		# open DEBUG, ">".$_->{engine}{name}."-".$_->{from}.".html" or die $!;
		# print DEBUG $_->{_content};
		# close DEBUG;
		#  ,  
		my ($results, $next_url) = 
		    ($_->{_header} =~ /^Location: (.*?)\r$/ms) ?
		    ([], abs_url($1, $_->{engine}{url})) :
		    ParseEngine($_->{_content}, $_->{engine});
		#       max_results
		my $from = $_->{from} + @$results;
		my $max = $engines{$_->{engine}{name}}{max_results};
		$URL{ $_->{engine}{name}." - ".$from } = {
			url	=> $next_url,
			engine	=> $_->{engine},
			from	=> $from,
		    } if $next_url and (not defined $max or $max >= $from);
		#   
		@{$engines{$_->{engine}{name}}{results}}[$_->{from}..$from-1] =
		    @$results;
		delete $URL{$_->{engine}{name}." - ".$_->{from}};
	    }
	    #  timeout
	    undef(%URL) if $timeout <= time();
	});
    #    max_results    0-  (undef)
    $_->{results} = defined($_->{max_results}) ?
	[ splice( @{$_->{results}}, 1, $_->{max_results}) ] :
	[ splice( @{$_->{results}}, 1) ]  for values %engines;
    return \%engines;
}

=back

=head1 SUPPORT (UPDATE SEARCH SERVERS)

 Bes Murzik <murzik@powerman.sky.net.ua>

=head1 AUTHORS

 Alex Efros <powerman@consultant.com>
 Nikita Savin <nikita@only.com>

=cut

1;
